#include <gsl/gsl_math.h>
#include <gsl/gsl_complex.h>
#include <gsl/gsl_complex_math.h>
#include <gsl/gsl_matrix_complex_double.h>
#include <gsl/gsl_fft_complex.h>
#include <gsl/gsl_fft_real.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_linalg.h>
#include <iostream>
#include <Qt/qimage.h>
#include <R2Shapes/R2Shapes.h>
#include <Qt/qimage.h>
#include "opencv/cv.h"
#include "opencv/highgui.h"


#ifndef __LINEAR_ALGEBRA_H
#define __LINEAR_ALGEBRA_H

class SymVector;
class SymMatrix2D;
typedef gsl_complex SymComplex;

enum EntryType
{
	SYM_ENTRY_REAL,
	SYM_ENTRY_COMPLEX
};

class SymVector
{
	public:
		virtual ~SymVector(){}	
		virtual EntryType GetEntryType() = 0;
		int m;	
		bool m_bFreeData;
};

class SymVectorComplex : public SymVector
{
	public:
		SymVectorComplex(int m);
		virtual ~SymVectorComplex();
	
		SymComplex & operator()(int i);
		virtual EntryType GetEntryType();
		
	protected:
		gsl_vector_complex * m_pData;
};

class SymVectorReal : public SymVector
{
	public:
		SymVectorReal(int m, double initVal);
		SymVectorReal(int m);
		SymVectorReal(gsl_vector * vec, bool shallowCopyOnly = true);
		SymVectorReal(QString inFile);
		SymVectorReal(std::vector<double> & inVector);
	
		void SaveSymVector(QString outFile);
		void LoadSymVector(FILE * inFile);
		void CopyValuesFrom(gsl_vector * vec);
		std::vector<double> ToVector();
		virtual ~SymVectorReal();
		double & operator()(int i);
		virtual EntryType GetEntryType();
		SymVectorReal & operator = (SymVectorReal & vec);
		double SumElements(double p);
		void Multiply(double val);
		void Add(SymVectorReal & v2);
		void SetAll(double val);
		
		void PrintVector(double absMin = 0);

		gsl_vector * m_pData;
};

class SymMatrix
{
	public:
		enum SVDMethod
		{
			SVD_AUTO,
			SVD_REGULAR, 
			SVD_GOLUB_REINSCH,		// use if M>>N
			SVD_JACOBI				// use if M>>N and higher accuracy is desired
		};
		enum InverseMethod
		{
			INV_AUTO,
			INV_LU,
			INV_SVD
		};
	
		SymMatrix(){}
		virtual ~SymMatrix(){}
		virtual int GetNumDims() = 0;
		virtual EntryType GetEntryType() = 0;
		virtual void PrintMatrix(double ){std::cout<<"SymMatrix: Print not implemented"<<std::endl;}
		virtual double SumElements(double){assert(false);}
		virtual int GetNumElements(){assert(false);}
		virtual void DivideBy(double){assert(false);}
		virtual void PointwiseMultiply(double){assert(false);}	
		virtual double GetMax(bool ){assert(false); return 0;}
		virtual double GetMin(bool ){assert(false);		return 0;}
		virtual void ConstMinusValue(double ){assert(false);}
		virtual void Subtract(double ){assert(false);}	
		virtual void Subtract(SymMatrix & ){assert(false);}
		bool m_bFreeData;
};

/**
 * This is like 3D matrix, where 3-rd dimension is (very) discrete and might not 
 * be correlated with neighboring dimensions
 */
class SymMatrixCollection : public SymMatrix
{
	public:
		SymMatrixCollection(std::vector<SymMatrix*> & matrices);
		SymMatrixCollection(int N1, int N2);
		void AddMatrixToCollection(SymMatrix* matrix, int id);
		void AddMatrixToCollection(int i, int j);
		SymMatrix & GetMatrixInCollection(int i);
		SymMatrix & GetMatrixInCollection(int i, int j);	
		int GetNumDims() = 0;
		EntryType GetEntryType() = 0;

};


class SymMatrix2D : public SymMatrix
{
	public:
		SymMatrix2D(int m, int n){this->m = m; this->n = n;}
		virtual ~SymMatrix2D(){}
		virtual void WriteRow(int row, gsl_complex_packed_array data) = 0;
		virtual void WriteCol(int col, gsl_complex_packed_array data) = 0;
		virtual void ReadRow(int row, gsl_complex_packed_array data) = 0;
		virtual void ReadCol(int col, gsl_complex_packed_array data) = 0;
		virtual float * GetRealDataArray() = 0;
		virtual QImage ToImage(bool shiftByMin = true) = 0;
		virtual void CopySquare(SymMatrix2D * , int , int , int , int , int , int) {assert(false);}

//		virtual void PointwisePower(double p) = 0;
//		virtual double SumElements() = 0;
//		virtual void DivideBy(double value) = 0;
	
		int m, n;	
};



class SymMatrix2DComplex : public SymMatrix2D
{
	public:
		SymMatrix2DComplex(int m, int n);
		virtual ~SymMatrix2DComplex();
		virtual void Clear();
	
		virtual int GetNumDims();
		virtual EntryType GetEntryType();	
		virtual void PrintMatrix(double absMax = 0 );
		virtual QImage ToImage(bool shiftByMin = true);
	
	//easy to use version
		SymMatrix2DComplex operator * (SymMatrix2DComplex & m2);
		virtual float * GetRealDataArray();
		virtual void WriteRow(int row, gsl_complex_packed_array data);
		virtual void WriteCol(int col, gsl_complex_packed_array data);
		virtual void ReadRow(int row, gsl_complex_packed_array data);
		virtual void ReadCol(int col, gsl_complex_packed_array data);
	
	// cheap version
		virtual void PointwiseMultiply(SymMatrix2DComplex & m2, SymMatrix2DComplex & trg);
		virtual void Multiply(SymMatrix2DComplex & m2, SymMatrix2DComplex & trg);
		virtual void Conj(SymMatrix2DComplex & trg);		
		virtual void Transpose();
		virtual void Transpose(SymMatrix2DComplex & trg);
		virtual double GetMax(bool absVal);
		virtual double GetMin(bool absVal);	
		virtual void DivideBy(double val);
//		virtual void PointwisePower(double p);
//		virtual double SumElements();
//		virtual void DivideBy(double value);
		SymMatrix2DComplex & operator=(SymMatrix2DComplex & src);	
	
		SymComplex & operator()(int i, int j);
		
		gsl_matrix_complex * m_pData;
};

class SymMatrix2DReal : public SymMatrix2D
{
	public:
		SymMatrix2DReal(int m, int n);
		SymMatrix2DReal(int m, int n, double initVal);	
		SymMatrix2DReal(QImage & img);	
		SymMatrix2DReal(SymMatrix2DReal & img);		
		virtual ~SymMatrix2DReal();
		virtual void Clear();
		virtual float * GetRealDataArray();
		
		int GetNumDims();
		virtual EntryType GetEntryType();
		virtual void PrintMatrix(double absMax=0);
		virtual QImage ToImage(bool shiftByMin = true);
		virtual QImage ClampToImage(double minVal, double maxVal);
		virtual void FromImage(QImage & img);
		
		//easy to use version
		SymMatrix2DReal operator * (SymMatrix2DReal & m2);
		virtual void WriteRow(int row, gsl_complex_packed_array data);
		virtual void WriteCol(int col, gsl_complex_packed_array data);
		virtual void ReadRow(int row, gsl_complex_packed_array data);
		virtual void ReadCol(int col, gsl_complex_packed_array data);
		
		// cheap version
		virtual void PointwiseMultiply(SymMatrix2DReal & m2, SymMatrix2DReal & trg);
		virtual void Multiply(SymMatrix2DReal & m2, SymMatrix2DReal & trg);
		virtual void Multiply(SymVectorReal & v, SymVectorReal & vTrg);
		virtual void Transpose(SymMatrix2DReal & trg);	
		virtual void Inverse(InverseMethod method = INV_LU, SymMatrix2DReal * workA = NULL);
		virtual void ThinSVD(SymMatrix2DReal * U, SymMatrix2DReal * V, SymVectorReal * S, SVDMethod = SVD_REGULAR, 
							 SymVectorReal * work = NULL, SymMatrix2DReal * workX = NULL);
		
		virtual void DecomposeCholesky();	
		virtual void SolveUsingCholesky(SymVectorReal & b, SymVectorReal &x);

		virtual void PointwisePower(double p, bool absVal);
		virtual double SumElements(double p);		//pth norm
		virtual double Mean();					//mean value
		virtual int GetNumElements();
		virtual void DivideBy(double value);
		virtual double GetMax(bool absVal);
		virtual double GetMin(bool absVal);	
		virtual void Subtract(double value);	
		virtual void Subtract(SymMatrix2DReal & otherMatrix);
		virtual void Subtract(SymMatrix & otherMatrix);		
		virtual void ConstMinusValue(double c);
		virtual void Add(SymMatrix2DReal & otherMatrix);		
		virtual double CalculateDifference(SymMatrix2DReal & m2, double p);
		virtual double SampleBilinear(double x, double y);
		virtual void CopySquare(SymMatrix2D * trg, int minSrcX, int minSrcY, int minTrgX, int minTrgY, int widthX, int widthY);
		SymMatrix2DReal & operator=(SymMatrix2DComplex & src);
		SymMatrix2DReal & operator=(SymMatrix2DReal & src);
		double & operator()(int i, int j);
		
		gsl_matrix * m_pData;
	
	//filters
		virtual void GaussBlur(double std_dev);	
		void FillGrid();
		void CopyFromGrid();
		R2Grid * m_pTempFilterGrid;
		
		IplImage * m_pTempFilterOpenCVImg32F;
		IplImage * m_pTempProcessOpenCVImg32F;
		IplImage * m_pTempFilterOpenCVImg8U;
		IplImage * m_pTempProcessOpenCVImg8U;	
	
		void InitOpenCV();
		void FillOpenCV(IplImage * trg, double scale = 1);
		void CopyFromOpenCV(IplImage * src, double scale = 1);
		void CannyFilterOpenCV(double threshold1, double threshold2, int window = 3);
		void SobelFilterOpenCV(int window=3);
		void BlurFilterOpenCV(double stdDev);
		
		
};

#endif
