#include "DeformImageWarpLinearMesh.h"
#include "CVLib.h"
#include "DrawingRoutines.h"
#include <iostream>
#include <time.h>

// TESTING Fns
QImage ImageWarp::TestCreateCheckerboard(int numCellsX, int numCellsY, int cellWidth, int cellHeight, double rndLevel, QString meshing)
{
	int width = cellWidth * numCellsX;
	int height = cellHeight * numCellsY;
	QImage chckbrd(width, height, QImage::Format_RGB32);
	
	for (int x=0; x<width; x++)
	{
		for (int y=0; y<height; y++)
		{
			int cellX = x / cellWidth;
			int cellY = y / cellHeight;
			QRgb color;
			if ((cellX+cellY)%2==0)
				color = qRgb(255, 255, 255);
			else
				color = qRgb(0,0,0);
			chckbrd.setPixel(x, y, color);
		}
	}
		 
	if (rndLevel>0)
	{
		int ctrlpnts[2];
		ctrlpnts[0] = numCellsX-1;
		ctrlpnts[1] = numCellsY-1;
		ImageWarpLinearMesh::FaceType meshingFaceType;
		if (meshing=="tri1")
			meshingFaceType = ImageWarpLinearMesh::FACE_TRIANGLE1;
		else if (meshing=="tri2")
			meshingFaceType = ImageWarpLinearMesh::FACE_TRIANGLE2;
		else if (meshing=="tri3")
			meshingFaceType = ImageWarpLinearMesh::FACE_TRIANGLE3;
		else if (meshing=="quad")
			meshingFaceType = ImageWarpLinearMesh::FACE_QUAD;
		
		ImageWarpLinearMesh deform(chckbrd, chckbrd.width(), chckbrd.height(), ctrlpnts, false, true, 1., 0, ImagePenalization(), NULL, meshingFaceType);
		deform.TestRandomizeParams(-rndLevel, rndLevel);
		chckbrd = deform.GetDeformedImage();
	}
	return chckbrd;
}

QImage ImageWarp::TestCreateRegularSpacedGlyph(int numGlyphX, int numGlyphY, R2Vector step1, R2Vector step2, 
											   double glyphWidth, double glyphHeight, QString glyphType)
{
	int width = (numGlyphX+1) * (ceil(glyphWidth + qMax(step1.X(), step2.X())));
	int height = (numGlyphY+1) * (ceil(glyphHeight + qMax(step1.Y(), step2.Y())));
	QImage img(width, height, QImage::Format_RGB32);
	QRgb white = qRgb(255, 255, 255);
	double boundaryX = qMax(step1.X(), step2.X())/2;
	double boundaryY = qMax(step1.Y(), step2.Y())/2;	
	img.fill(0);
//	std::cout<<"Step1=["<<step1.X()<<", "<<step1.Y()<<"] step2=["<<step2.X()<<", "<<step2.Y()<<"]"<<std::endl;
	for (int i=-20*numGlyphX; i<20*numGlyphX; i++)
	{
		for (int j=-20*numGlyphY; j<20*numGlyphY; j++)
		{
			double refX = boundaryX + (step1*i + step2*j).X();
			double refY = boundaryY + (step1*i + step2*j).Y();
			
			if (glyphType=="ellipse")
			{
				double centerX = refX+(double)glyphWidth/2.;
				double centerY = refY+(double)glyphHeight/2.;
				double radW = (double)glyphWidth/2.;
				double radH = (double)glyphHeight/2.;				
				
				for (int x=refX; x<refX+glyphWidth; x++)
					for (int y=refY; y<refY+glyphHeight; y++)
						if (pow((x-centerX)/radW, 2) + pow((y-centerY)/radH, 2)<=1
							&& x<img.width() && x>=0 && y<img.height() && y>=0)
							img.setPixel(x, y, white);				
			}
			else if (glyphType=="rectangle")
			{
				for (int x=refX; x<refX+glyphWidth; x++)
					for (int y=refY; y<refY+glyphHeight; y++)
						if (x<img.width() && x>=0 && y<img.height() && y>=0)
							img.setPixel(x, y, white);
			}
			else if (glyphType=="cross")
			{
				int lineStep = qMin(glyphWidth/10, glyphHeight/10);
				bool horizontalScanline = glyphWidth > glyphHeight;
				
				for (int x=0; x<qMax(glyphWidth-1, glyphHeight); x++)
				{
					int locX = refX+x;
					int locY = refY+x;
					if (horizontalScanline)
						locY = refY+((double)x/(double)(glyphWidth)) * glyphHeight;
					else
						locX = refX+((double)x/(double)(glyphHeight)) * glyphWidth;
					
					for (int lx = locX-lineStep; lx<=locX+lineStep; lx++)
					{
						if (lx<img.width() && lx>=0 && locY<img.height() && locY>=0)
							img.setPixel(lx, locY, white);
						if (2*refX - lx + glyphWidth<img.width() && 2*refX - lx + glyphWidth>=0 
							&& locY<img.height() && locY>=0)						
							img.setPixel(2*refX - lx + glyphWidth, locY, white);
					}
				}
			}
		}
	}
	return img;
}


ImageWarpTestExample::ImageWarpTestExample(QString exampleName)
: ExampleWithAnswer(exampleName)
{	
	m_BackgroundColor = 0;	
	m_NumX = 8;
	m_NumY = 8;
	m_Blur = 0;
	m_V1 = 15;
	m_V2 = 15;
	m_OddShift = 0;
	
	//glyphs
	m_GlyphName = "ellipse";
	m_GlyphX = 5;
	m_GlyphY = 5;
	m_GlyphMeanColor = 1;
	m_bPutCtrlAtGlyphCorners = false;	
	m_bHierarchical = false;
	
	//noise params
	m_FractionMissing = 0.1;					// 0 to 1
	m_FractionExtra = 0.1;						// 0 to 1
	m_ExtraBlobs = 2;							// any integer > 0
	m_ExtraBlobsFractionArea = .1;			// 0 to 1
	m_GlyphLocationVariationHighFreq = 10;	// O(V1 or V2)
	m_GlyphLocationVariationLowFreq = 8;		// O(V1 or V2)
	m_GlyphDimensionsVariation = 5;			// O(glyphX or glyphY)
	m_GlyphColorVariations = 5;				// 0 to 2
	m_GlyphQuadVariations = 0;	
}

void ImageWarpTestExample::PrintParams()
{
	std::cout<<"Lattice:"<<std::endl;
	std::cout<<"\tBackground Color = "<<m_BackgroundColor<<std::endl;
	std::cout<<"\tNum Points=["<<m_NumX<<", "<<m_NumY<<"]"<<std::endl;
	std::cout<<"\tBlur="<<m_Blur<<std::endl;	
	std::cout<<"\tGenerator Vectors=["<<m_V1.X()<<", "<<m_V1.Y()<<"] x ["<<m_V2.X()<<", "<<m_V2.Y()<<"]"<<std::endl;
	std::cout<<"\tOddShift="<<m_OddShift<<std::endl;
	std::cout<<"Glyph:"<<std::endl;	
	std::cout<<"\tName="<<m_GlyphName.toStdString().c_str()<<std::endl;
	std::cout<<"\tDim= "<<m_GlyphX<<" x "<<m_GlyphY<<std::endl;	
	std::cout<<"\tColor= "<<m_GlyphMeanColor<<std::endl;
	std::cout<<"\tPutControlsAtCorners="<<(m_bPutCtrlAtGlyphCorners ? "true" : "false")<<std::endl;
	std::cout<<"Noise:"<<std::endl;		
	std::cout<<"\tFraction Missing = "<<m_FractionMissing<<std::endl;
	std::cout<<"\tFraction Extra = "<<m_FractionExtra<<std::endl;	
	std::cout<<"\tBlobs: "<<m_ExtraBlobs<<" with area "<<m_ExtraBlobsFractionArea<<std::endl;	
	std::cout<<"\tLoc. Noise: high freq.="<<m_GlyphLocationVariationHighFreq<<" low freq.="<<m_GlyphLocationVariationLowFreq<<std::endl;		
	std::cout<<"\tColor Variation = "<<m_GlyphColorVariations<<std::endl;
	std::cout<<"\tDimNoise = "<<m_GlyphDimensionsVariation<<std::endl;
	std::cout<<"\tQuadNoise = "<<m_GlyphQuadVariations<<std::endl;	
}

void ImageWarpTestExample::GenerateExample(int randomSeed)
{
	PrintParams();
	std::cout<<"\tRandomSeed="<<randomSeed<<std::endl;
	srand(randomSeed);
	int width = (int)((m_NumX) * m_V1.X() + (m_NumY) * m_V2.X()) - 1;
	int height = (int)((m_NumX) * m_V1.Y() + (m_NumY) * m_V2.Y()) - 1;
	double offsetX = (m_V1.X() + m_V2.X())/2;
	double offsetY = (m_V1.Y() + m_V2.Y())/2;	
	m_ImageInputExample = QImage(width, height, QImage::Format_RGB32);
	m_ImagePerfectOutput = QImage(width, height, QImage::Format_RGB32);

//	std::cout<<"Background color="<<m_BackgroundColor<<std::endl;
	DrawingRoutines::ClearImage(m_ImageInputExample, qRgb((int)(m_BackgroundColor*255), (int)(m_BackgroundColor*255), (int)(m_BackgroundColor*255)));
	DrawingRoutines::ClearImage(m_ImagePerfectOutput, qRgb((int)(m_BackgroundColor*255), (int)(m_BackgroundColor*255), (int)(m_BackgroundColor*255)));	

	double coeffA = m_GlyphLocationVariationLowFreq * (double)(rand()%100000) / 100000.;
	double coeffB = m_GlyphLocationVariationLowFreq * (double)(rand()%100000) / 100000.;
	double coeffC = m_GlyphLocationVariationLowFreq * (double)(rand()%100000) / 100000.;
	double coeffD = m_GlyphLocationVariationLowFreq * (double)(rand()%100000) / 100000.;
	double rndA = (double)(rand()%100000) / 100000.;
	double rndB = 4*(double)(rand()%100000) / 100000.;	
	double rndC = (double)(rand()%100000) / 100000.;
	double rndD = 4*(double)(rand()%100000) / 100000.;		
	double freqA = ((3+rndA)*3.14159265) / (2*width);//(width * (1 + );
	double freqB = ((4+rndB)*3.14159265) / (2*width);//(width * (1 + (double)(rand()%100000) / 100000.));
	double freqC = ((3+rndC)*3.14159265) / (2*height);//(height * (1 + (double)(rand()%100000) / 100000.));
	double freqD = ((4+rndD)*3.14159265) / (2*height);//(height * (1 + (double)(rand()%100000) / 100000.));
	
	bool * vertexInPixel = new bool[(width+1) * (height+1)];
	
	for (int i=0; i<(width+1) * (height+1); i++)
		vertexInPixel[i] = false;	
	
	std::vector<double> featurePnts;
	std::vector<double> properLocations;
	
	// put glyphs on lattice
	for (int i=0; i<m_NumX; i++)
	{
		for (int j=0; j<m_NumY; j++)
		{			
			if (j%2==1)
				offsetX = offsetX + m_OddShift;
			else if (j>0)
				offsetX = offsetX - m_OddShift;
			
			if (((double)(rand()%100000)) / 100000. < m_FractionMissing)
				continue;
							
			double glyphLocX = offsetX + i * m_V1.X() + j * m_V2.X() + 0.5;
			double glyphLocY = offsetY + i * m_V1.Y() + j * m_V2.Y() + 0.5;
			
			double highFreqLocNoiseX = ((double)(rand()%100000)) / 100000. * m_GlyphLocationVariationHighFreq - m_GlyphLocationVariationHighFreq/2;
			double highFreqLocNoiseY = ((double)(rand()%100000)) / 100000. * m_GlyphLocationVariationHighFreq - m_GlyphLocationVariationHighFreq/2;
			double lowFreqLocNoiseX = coeffA * sin(glyphLocX * freqA) + coeffB * sin(glyphLocY * freqB);
			double lowFreqLocNoiseY = coeffC * sin(glyphLocX * freqC) + coeffD * sin(glyphLocY * freqD);
			if (i<m_NumX/2 && j<m_NumY/2)
			{
				lowFreqLocNoiseX = 0;
				lowFreqLocNoiseY = 0;
			}

			double glyphColor = m_GlyphMeanColor + m_GlyphColorVariations * ((double)(rand()%100000)) / 100000. - m_GlyphColorVariations/2.;
			glyphColor = qMax(qMin(1., glyphColor), 0.);
//			std::cout<<"1 Glyph color="<<glyphColor<<std::endl;						
			double glyphWidth = m_GlyphX + m_GlyphDimensionsVariation * ((double)(rand()%100000)) / 100000. - m_GlyphDimensionsVariation/2;
			glyphWidth = qMax(glyphWidth, 0.);
			double glyphHeight = m_GlyphY + m_GlyphDimensionsVariation * ((double)(rand()%100000)) / 100000. - m_GlyphDimensionsVariation/2;			
			glyphHeight = qMax(glyphHeight, 0.);
			
			double glyphLocWithNoiseX = glyphLocX + highFreqLocNoiseX + lowFreqLocNoiseX;
			double glyphLocWithNoiseY = glyphLocY + highFreqLocNoiseY + lowFreqLocNoiseY;			

			if (!m_bPutCtrlAtGlyphCorners)
			{				
				if ((int)glyphLocWithNoiseY < width && (int)glyphLocWithNoiseX<height
					&& vertexInPixel[((int)glyphLocWithNoiseY) * (width+1) + ((int)glyphLocWithNoiseX)]==false
					&& vertexInPixel[((int)glyphLocY) * (width+1) + ((int)glyphLocX)]==false)
				{
					vertexInPixel[((int)glyphLocWithNoiseY) * (width+1) + ((int)glyphLocWithNoiseX)] = true;
					vertexInPixel[((int)glyphLocY) * (width+1) + ((int)glyphLocX)]= true;
					featurePnts.push_back(glyphLocWithNoiseX);
					featurePnts.push_back(glyphLocWithNoiseY);			
					properLocations.push_back(glyphLocX);
					properLocations.push_back(glyphLocY);
				}
			}
			else
			{				
				double newGlyphLocX = glyphLocWithNoiseX - glyphWidth/2;
				double newGlyphLocY = glyphLocWithNoiseY - glyphHeight/2;				
				if ((int)newGlyphLocX <= width && (int)newGlyphLocY<= height
					&& vertexInPixel[((int)newGlyphLocY) * (width+1) + ((int)newGlyphLocX)]==false
					&& vertexInPixel[((int)(glyphLocY - m_GlyphY/2)) * (width+1) + ((int)(glyphLocX - m_GlyphX/2))]==false)
				{
					vertexInPixel[((int)newGlyphLocY) * (width+1) + ((int)newGlyphLocX)] = true;
					vertexInPixel[((int)(glyphLocY - m_GlyphY/2)) * (width+1) + ((int)(glyphLocX - m_GlyphX/2))]= true;
					featurePnts.push_back(newGlyphLocX);
					featurePnts.push_back(newGlyphLocY);			
					properLocations.push_back(glyphLocX - m_GlyphX/2);
					properLocations.push_back(glyphLocY - m_GlyphY/2);
				}

				newGlyphLocX = glyphLocWithNoiseX + glyphWidth/2;
				newGlyphLocY = glyphLocWithNoiseY + glyphHeight/2;				
				if ((int)newGlyphLocX <=width && (int)newGlyphLocY<=height
					&& vertexInPixel[((int)newGlyphLocY) * (width+1) + ((int)newGlyphLocX)]==false
					&& vertexInPixel[((int)(glyphLocY + m_GlyphY/2)) * (width+1) + ((int)(glyphLocX + m_GlyphX/2))]==false)
				{
					vertexInPixel[((int)newGlyphLocY) * (width+1) + ((int)newGlyphLocX)] = true;
					vertexInPixel[((int)(glyphLocY + m_GlyphY/2)) * (width+1) + ((int)(glyphLocX + m_GlyphX/2))]= true;				
					featurePnts.push_back(newGlyphLocX);
					featurePnts.push_back(newGlyphLocY);			
					properLocations.push_back(glyphLocX + m_GlyphX/2);
					properLocations.push_back(glyphLocY + m_GlyphY/2);
				}

				newGlyphLocX = glyphLocWithNoiseX + glyphWidth/2;
				newGlyphLocY = glyphLocWithNoiseY - glyphHeight/2;				
				if ((int)newGlyphLocX <=width && (int)newGlyphLocY<=height
					&& vertexInPixel[((int)newGlyphLocY) * (width+1) + ((int)newGlyphLocX)]==false
					&& vertexInPixel[((int)(glyphLocY + m_GlyphY/2)) * (width+1) + ((int)(glyphLocX - m_GlyphX/2))]==false)
				{
					vertexInPixel[((int)newGlyphLocY) * (width+1) + ((int)newGlyphLocX)] = true;
					vertexInPixel[((int)(glyphLocY + m_GlyphY/2)) * (width+1) + ((int)(glyphLocX - m_GlyphX/2))]= true;										
					featurePnts.push_back(newGlyphLocX);
					featurePnts.push_back(newGlyphLocY);			
					properLocations.push_back(glyphLocX + m_GlyphX/2);
					properLocations.push_back(glyphLocY - m_GlyphY/2);
				}

				newGlyphLocX = glyphLocWithNoiseX - glyphWidth/2;
				newGlyphLocY = glyphLocWithNoiseY + glyphHeight/2;				
				if ((int)newGlyphLocX <= width && (int)newGlyphLocY<=height
					&& vertexInPixel[((int)newGlyphLocY) * (width+1) + ((int)newGlyphLocX)]==false
					&& vertexInPixel[((int)(glyphLocY + m_GlyphY/2)) * (width+1) + ((int)(glyphLocX - m_GlyphX/2))]==false)
				{
					vertexInPixel[((int)newGlyphLocY) * (width+1) + ((int)newGlyphLocX)] = true;
					vertexInPixel[((int)(glyphLocY + m_GlyphY/2)) * (width+1) + ((int)(glyphLocX - m_GlyphX/2))]= true;										
					featurePnts.push_back(newGlyphLocX);
					featurePnts.push_back(newGlyphLocY);			
					properLocations.push_back(glyphLocX - m_GlyphX/2);
					properLocations.push_back(glyphLocY + m_GlyphY/2);
				}				
			}
			
			AddGlyph(m_ImageInputExample, glyphLocWithNoiseX, glyphLocWithNoiseY, glyphColor, glyphWidth, glyphHeight);
			AddGlyph(m_ImagePerfectOutput, glyphLocX, glyphLocY, glyphColor, m_GlyphX, m_GlyphY);
		}
	}
	
	int numRigidControls = featurePnts.size()/2;

	std::vector<double> savedFeaturePnts = featurePnts;
	// sort vertices
	if (m_bHierarchical)
	{
		std::vector<int> triangles;		
		CVLibRoutines::GetDelaunayTriangulation(featurePnts, triangles);
		std::vector<int> vertexOrder;
		CVLibRoutines::SortVerticesByDecimation(m_ImageInputExample, featurePnts, triangles, vertexOrder);

	//	for (int i=0; i<featurePnts.size()/2; i++)
	//		vertexOrder.push_back(i);
//		for (int i=0; i<(int)vertexOrder.size(); i++)
//		{
//			//std::cout<<"Swapping: "<<i<<" <-> "<<vertexOrder[i]<<std::endl;
//			//swap vertexOrder[i] and i
//			double valX = properLocations[2*i+0];
//			double valY = properLocations[2*i+1];
//			
//			properLocations[2*i+0] = properLocations[2*vertexOrder[i]+0];
//			properLocations[2*i+1] = properLocations[2*vertexOrder[i]+1];
//			
//			properLocations[2*vertexOrder[i]+0] = valX;
//			properLocations[2*vertexOrder[i]+1] = valY;		
//		}
	}
	
	//add some regular lattice
	double stepDX = qMax(m_V1.X(), m_V2.X());
	double stepDY = qMax(m_V1.Y(), m_V2.Y());	
	int stepsX = width / stepDX;
	int stepsY = height / stepDY;	

	int newLatticeX = stepsX * 2;
	int newLatticeY = stepsY * 2;	
	for (int i=0; i<newLatticeX; i++)
	{
		for (int j=0; j<newLatticeY; j++)
		{
			double xLoc = ((double)i*(double)width / (double)(newLatticeX-1));
			double yLoc = ((double)j*(double)height / (double)(newLatticeY-1));
			if (!vertexInPixel[((int)yLoc) * (width+1) + ((int)xLoc)])
			{
				featurePnts.push_back(xLoc);
				featurePnts.push_back(yLoc);
				savedFeaturePnts.push_back(xLoc);
				savedFeaturePnts.push_back(yLoc);
			}
		}
	}

	// add extra glyphs
	for (int i=0; i<(int)(m_FractionExtra * (double)(m_NumX * m_NumY)); i++)
	{
		double glyphLocX = ((double)(rand()%100000)) / 100000. * width;
		double glyphLocY = ((double)(rand()%100000)) / 100000. * height;
		double glyphColor = m_GlyphMeanColor + m_GlyphColorVariations * ((double)(rand()%100000)) / 100000.;
		glyphColor = qMax(qMin(1., glyphColor), 0.);		
//		std::cout<<"2. glyphColor="<<glyphColor<<std::endl;		
		double glyphWidth = m_GlyphX + m_GlyphDimensionsVariation * ((double)(rand()%100000)) / 100000.;
		double glyphHeight = m_GlyphY + m_GlyphDimensionsVariation * ((double)(rand()%100000)) / 100000.;			
		glyphWidth = qMax(glyphWidth, 0.);
		glyphHeight = qMax(glyphHeight, 0.);
		
		AddGlyph(m_ImageInputExample, glyphLocX, glyphLocY, glyphColor, glyphWidth, glyphHeight);
		AddGlyph(m_ImagePerfectOutput, glyphLocX, glyphLocY, glyphColor, glyphWidth, glyphHeight);
	}
	// add extra blobs
	double area = width * height;
	for (int i=0; i<m_ExtraBlobs; i++)
	{
		double blobX = ((double)(rand()%100000)) / 100000.;
		double blobY = ((double)(rand()%100000)) / 100000.;
		int blobColor = (int)(((double)(rand()%100000)) / 100000. * 255.);
//		std::cout<<"blobColor="<<blobColor<<std::endl;
		
		DrawingRoutines::DrawCircle(m_ImageInputExample, blobX, blobY, 
									sqrt(m_ExtraBlobsFractionArea * area/3.14159265) / qMax(width, height), 0,
									qRgb(blobColor, blobColor, blobColor), qRgb(0,0,0));
		DrawingRoutines::DrawCircle(m_ImagePerfectOutput, blobX, blobY, 
									sqrt(m_ExtraBlobsFractionArea * area/3.14159265)  / qMax(width, height), 0,
									qRgb(blobColor, blobColor, blobColor), qRgb(0,0,0));
	}

	std::cout<<"numRigidControls="<<numRigidControls<<std::endl;	
//	if (!m_bHierarchical)
//	{
		m_pPerfectSolution = new SymVectorReal(properLocations.size());
		for (int i=0; i<(int)properLocations.size(); i++)
			(*m_pPerfectSolution)(i) = properLocations[i];
//	}

	m_pValidSolutionSpace = CreateDeformationModel(m_ImageInputExample, featurePnts, numRigidControls);
	m_pValidSolutionSpace->SwitchToHierarchy(0);
//	((ImageWarpLinearMesh*)m_pValidSolutionSpace)->m_bShowWireframe = true;
//	((ImageWarpLinearMesh*)m_pValidSolutionSpace)->DrawGeodesicSorting().save("DeformGeodesicSorting.png");

	// Create 'perfect' solution
	ImageWarpLinearMesh * tempSolution = CreateDeformationModel(m_ImageInputExample, savedFeaturePnts, numRigidControls);
	tempSolution->SwitchToHierarchy(0);
	tempSolution->DeformAtCurrentHierarchyLevel(*m_pPerfectSolution);
	m_ImagePerfectDeformed = tempSolution->GetDeformedImage();
}

ImageWarpLinearMesh * ImageWarpTestExample::CreateDeformationModel(QImage & inImg, std::vector<double> & featurePnts, int numRigidControls)
{
	//triangulate
	std::vector<int> triangles;
	int numPnts = featurePnts.size()/2;
	triangles.clear();
	CVLibRoutines::FindFeatureMesh(m_ImageInputExample, numPnts, featurePnts, triangles);
	
	//initialize vertices
	std::vector<ImageWarp::ImgVertex> imgVertices;
	for (int i=0; i<(int)featurePnts.size(); i+=2)
	{
		ImageWarp::ImgVertex v;
		v.x = featurePnts[i+0];		v.y = featurePnts[i+1];		
		v.tu = featurePnts[i+0];	v.tv = featurePnts[i+1];		
		
		v.m_BoundaryVertexFlags = ImageWarp::ImgVertex::BND_VERTEX_NOT_FIXED;
		if (v.x==0 || v.x==inImg.width())
		{
			v.m_BoundaryVertexFlags |= ImageWarp::ImgVertex::BND_VERTEX_FIXED_X;
			v.m_fBoundaryX = v.x;
		}
		if (v.y==0 || v.y==inImg.height())
		{
			v.m_BoundaryVertexFlags |= ImageWarp::ImgVertex::BND_VERTEX_FIXED_Y;				
			v.m_fBoundaryY = v.y;
		}
		
		v.m_iParamID = i/2;
		imgVertices.push_back(v);
	}
	
	//initialize faces
	std::vector<ImageWarp::ImgFace> imgFaces;
	for (int i=0; i<(int)triangles.size(); i+=3)
	{
		ImageWarp::ImgFace f;
		f.v1 = triangles[i+0];		f.v2 = triangles[i+1];		f.v3 = triangles[i+2];			f.v4 = -1;
		imgFaces.push_back(f);
	}
	
	// Create deformation model
	ImageWarp::ImageHierarchyDescriptor * hierarchyDescr = new ImageWarp::ImageHierarchyDescriptor();
	hierarchyDescr->m_HierarchyType = ImageWarp::AS_RIGID_AS_POSSIBLE;
	if (m_bHierarchical)
	{
		hierarchyDescr->m_iMinPnts = 4;
		hierarchyDescr->m_iMaxPnts = numRigidControls;
	}
	else
	{
		hierarchyDescr->m_iMinPnts = numRigidControls;
		hierarchyDescr->m_iMaxPnts = numRigidControls;
	}
	double step = qMax(qMax(m_GlyphLocationVariationHighFreq / 4, m_GlyphLocationVariationLowFreq / 4), 2.);
	return new ImageWarpLinearMesh(inImg, inImg.width(), inImg.height(), imgVertices, imgFaces, 
									false, true, step, 1, ImageWarp::ImagePenalization(), 
								   hierarchyDescr);	
}

void ImageWarpTestExample::AddGlyph(QImage & img, double locX, double locY, double color, double dimX, double dimY)
{
	//glyphs
	if (m_GlyphName.startsWith("ellipse"))
	{
		DrawingRoutines::DrawEllipse(img, locX / img.width(), locY/img.height(), dimX/(2*img.width()), dimY/(2*img.height()),
					qRgb((int)(color*255), (int)(color*255), (int)(color*255)));
	}
	else if (m_GlyphName.startsWith("rectangle"))
	{
		DrawingRoutines::DrawRectangle(img, (locX-dimX/2)/img.width(), (locY-dimY/2)/img.height(), 
									   dimX/img.width(), dimY/img.height(), 
									   qRgb((int)(color*255), (int)(color*255), (int)(color*255)));
	}
	else if (m_GlyphName.startsWith("cross"))
	{		
		double width = qMax(qMin(dimX, dimY) / 10., 1.) / qMin(img.width(), img.height());
		
		DrawingRoutines::DrawLine(img, (locX-dimX/2)/img.width(), (locY-dimY/2)/img.height(), 
								  (locX+dimX/2)/img.width(), (locY+dimY/2)/img.height(),
								  width, qRgb((int)(color*255), (int)(color*255), (int)(color*255)));
		DrawingRoutines::DrawLine(img, (locX+dimX/2)/img.width(), (locY-dimY/2)/img.height(), 
								  (locX-dimX/2)/img.width(), (locY+dimY/2)/img.height(),
								  width, qRgb((int)(color*255), (int)(color*255), (int)(color*255)));
	}	
}

ImageWarpExampleCheckerboard::ImageWarpExampleCheckerboard(QString name)
: ExampleWithAnswer(name) 
{
}

void ImageWarpExampleCheckerboard::GenerateExample(int randomSeed)
{
	srand(randomSeed);
	QImage img = ImageWarp::TestCreateCheckerboard(m_NumX, m_NumY, m_WidthX, m_WidthY, 0);
	m_ImagePerfectOutput = img;
	
	int numPnts[2];
	numPnts[0] = m_NumX+1;
	numPnts[1] = m_NumY+1;	
	double rndLevel = m_fNoise;
	
	//create image
	ImageWarpLinearMesh deformedImg(img, img.width(), img.height(), numPnts, false, true, 1, 0, ImageWarpLinearMesh::ImagePenalization(), 
									NULL, ImageWarpLinearMesh::FACE_TRIANGLE3);
	SymVectorReal & paramsOriginal = deformedImg.GetParameterVector();
	for (int i=0; i<paramsOriginal.m; i++)
	{
		if (paramsOriginal(i)!=0 && paramsOriginal(i)!=img.width() && paramsOriginal(i)!=img.height())
			paramsOriginal(i) += ((rand()%1000000) * rndLevel / 1000000 - rndLevel/2);
	}
	deformedImg.m_bShowWireframe = false;
	((SymMatrix2DReal &) deformedImg.DeformAtCurrentHierarchyLevel(paramsOriginal));
	m_ImageInputExample = deformedImg.GetDeformedImage(0);		
	
	m_pPerfectSolution = new SymVectorReal(deformedImg.GetParameterVector());
	
	//create best
	ImageWarpLinearMesh deformPerfect(m_ImageInputExample, m_ImageInputExample.width(), m_ImageInputExample.height(), numPnts, true, true, 1, 0, ImageWarpLinearMesh::ImagePenalization(), 
									  NULL, ImageWarpLinearMesh::FACE_TRIANGLE3);
	
	((SymMatrix2DReal &) deformPerfect.DeformAtCurrentHierarchyLevel(*m_pPerfectSolution));	
	m_ImagePerfectDeformed = deformPerfect.GetDeformedImage(-1);
		
	//	//optimize
	m_pValidSolutionSpace = new ImageWarpLinearMesh(m_ImageInputExample, m_ImageInputExample.width(), m_ImageInputExample.height(), numPnts, true, true, 2, 0.7, ImageWarpLinearMesh::ImagePenalization(), 
													NULL, ImageWarpLinearMesh::FACE_TRIANGLE3);		
}

void ImageWarpExampleCheckerboard::PrintParams()
{
	std::cout<<"Checkerboard: Lattice="<<m_NumX<<" x "<<m_NumY<<" Dims="<<m_WidthX<<" x "<<m_WidthY<<" noise="<<m_fNoise<<std::endl;
}

void ImageWarpLinearMesh::TestImageWarpMeshing()
{
	//draw undeformed/deformed copies of image (with SAME slight deformations over regular grid)
	//no hierarchies, enable/disable wireframes
	
	//prepare image
	QImage img = TestCreateCheckerboard(8, 8, 50, 30, 0);
	
	int numPnts[2];
	numPnts[0] = 5;
	numPnts[1] = 5;	
	bool inv = false;
	bool fixBnd = true;
	double rndLevel = 25;
	if (inv)
		rndLevel = rndLevel/qMax(img.width(), img.height());
	bool wireframe = true;
	
	//Quads
	ImageWarpLinearMesh deformQuad(img, img.width(), img.height(), numPnts, inv, fixBnd, 1, 0, ImagePenalization(), NULL, FACE_QUAD);
	deformQuad.m_bShowWireframe = wireframe;
	deformQuad.TestRandomizeParams(-rndLevel, rndLevel);
	((SymMatrix2DReal &) deformQuad.DeformAtCurrentHierarchyLevel(deformQuad.GetParameterVector())).ToImage().save("ImageWarpQuadF.png");
	deformQuad.GetDeformedImage().save("ImgWarpQuad.png");
	
	//Triangle 1
	ImageWarpLinearMesh deformTri1(img, img.width(), img.height(), numPnts, inv, fixBnd, 1, 0, ImagePenalization(), NULL, FACE_TRIANGLE1);
	deformTri1.m_bShowWireframe = wireframe;
	((SymMatrix2DReal &) deformTri1.DeformAtCurrentHierarchyLevel(deformQuad.GetParameterVector())).ToImage().save("ImageWarpTri1F.png");
	deformTri1.GetDeformedImage().save("ImgWarpTri1.png");	
	
	//Triangle 2
	ImageWarpLinearMesh deformTri2(img, img.width(), img.height(), numPnts, inv, fixBnd, 1, 0, ImagePenalization(), NULL, FACE_TRIANGLE2);
	deformTri2.m_bShowWireframe = wireframe;
	((SymMatrix2DReal &) deformTri2.Deform(deformQuad.GetParameterVector())).ToImage().save("ImageWarpTri2F.png");
	deformTri2.GetDeformedImage().save("ImgWarpTri2.png");

	//Triangle 3
	ImageWarpLinearMesh deformTri3(img, img.width(), img.height(), numPnts, inv, fixBnd, 1, 0, ImagePenalization(), NULL, FACE_TRIANGLE3);
	deformTri3.m_bShowWireframe = wireframe;
	((SymMatrix2DReal &) deformTri3.DeformAtCurrentHierarchyLevel(deformQuad.GetParameterVector())).ToImage().save("ImageWarpTri3F.png");
	deformTri3.GetDeformedImage().save("ImgWarpTri3.png");
	
}

void ImageWarpLinearMesh::TestImageWarpHierarchies()
{
	//create deformation model with simple image
	//prepare image
	QImage img = TestCreateCheckerboard(8, 8, 50, 30, 0);
	
	int numPnts[2];
	numPnts[0] = 8;
	numPnts[1] = 8;
	bool inv = false;
	double rndLevelLow = 25;
	double rndLevelMed = 25;
	double rndLevelHigh = 15;
	if (inv)
		rndLevelHigh = rndLevelHigh/qMax(img.width(), img.height());
	
	bool wireframe = true;
	
	int highResDim = 8;
	int midResDim = 5;
	int lowResDim = 3;	
	
	ImageHierarchyDescriptor hierarchyDescr;
	hierarchyDescr.m_HierarchyType = OVERLAP_CHILDREN;
	hierarchyDescr.m_iAssocX = 4;
	hierarchyDescr.m_iAssocY = 4;
	hierarchyDescr.m_iOverlapX = 2;
	hierarchyDescr.m_iOverlapY = 2;	
	hierarchyDescr.m_iMinPnts = 1;
	hierarchyDescr.m_iInitNumPntsX = numPnts[0];
	hierarchyDescr.m_iInitNumPntsY = numPnts[1];	
	
	ImageWarpLinearMesh deformQuad(img, img.width(), img.height(), numPnts, inv, true, 1, 0, ImagePenalization(), &hierarchyDescr, FACE_QUAD);
	deformQuad.m_bShowWireframe = wireframe;
	
	SymVectorReal & params = deformQuad.GetParameterVector();
	
	//deform deform image low-level hierarchy
	int targetLowLevel = highResDim*highResDim*2 + midResDim*midResDim*2 + ((lowResDim-2)*lowResDim + (lowResDim-2))*2;
	params(targetLowLevel) += rndLevelLow;	params(targetLowLevel+1) += rndLevelLow;	
	((SymMatrix2DReal &) deformQuad.DeformAtCurrentHierarchyLevel(params)).ToImage().save("ImageWarp1LowF.png");
	deformQuad.GetDeformedImage().save("ImgWarp1Low.png");
	params(targetLowLevel) -= rndLevelLow;	params(targetLowLevel+1) -= rndLevelLow;		
	
	//deform deform image mid-level hierarchy
	int targetMidLevel = highResDim*highResDim*2 + ((midResDim-2)*midResDim + (midResDim-2))*2;
	params(targetMidLevel) += rndLevelMed;		params(targetMidLevel+1) += rndLevelMed;	
	((SymMatrix2DReal &) deformQuad.DeformAtCurrentHierarchyLevel(params)).ToImage().save("ImageWarp2MedF.png");
	deformQuad.GetDeformedImage().save("ImgWarp2Med.png");
	params(targetMidLevel) -= rndLevelMed;		params(targetMidLevel+1) -= rndLevelMed;
	
	//deform deform image top-level hierarchy	
	int targetHighLevel = ((highResDim-2)*highResDim + (highResDim-2))*2;
	params(targetHighLevel) += rndLevelHigh;		params(targetHighLevel+1) += rndLevelHigh;	
	((SymMatrix2DReal &) deformQuad.DeformAtCurrentHierarchyLevel(params)).ToImage().save("ImageWarp3HighF.png");
	deformQuad.GetDeformedImage().save("ImgWarp3High.png");
	params(targetHighLevel) -= rndLevelHigh;		params(targetHighLevel+1) -= rndLevelHigh;	

	//deform combined
	params(targetHighLevel) += rndLevelHigh;	params(targetHighLevel+1) += rndLevelHigh;	
	params(targetMidLevel) += rndLevelMed;		params(targetMidLevel+1) += rndLevelMed;	
	params(targetLowLevel) += rndLevelLow;		params(targetLowLevel+1) += rndLevelLow;	
	((SymMatrix2DReal &) deformQuad.DeformAtCurrentHierarchyLevel(params)).ToImage().save("ImageWarp4CombF.png");
	deformQuad.GetDeformedImage().save("ImgWarp4Comb.png");
}

void ImageWarpLinearMesh::TestImageWarpHierarchiesUsingSwitch()
{
	//create deformation model with simple image
	//prepare image
	QImage img = TestCreateCheckerboard(8, 8, 50, 30, 0);
	
	int numPnts[2];
	numPnts[0] = 8;
	numPnts[1] = 8;
	bool inv = false;
	double rndLevelLow = 25;
	double rndLevelMed = 25;
	double rndLevelHigh = 15;
	
	bool wireframe = true;
	
	int highResDim = 8;
	int midResDim = 5;
	int lowResDim = 3;	
	
	ImageHierarchyDescriptor hierarchyDescr;
	hierarchyDescr.m_HierarchyType = OVERLAP_CHILDREN;
	hierarchyDescr.m_iAssocX = 4;
	hierarchyDescr.m_iAssocY = 4;
	hierarchyDescr.m_iOverlapX = 2;
	hierarchyDescr.m_iOverlapY = 2;	
	hierarchyDescr.m_iMinPnts = 1;
	hierarchyDescr.m_iInitNumPntsX = numPnts[0];
	hierarchyDescr.m_iInitNumPntsY = numPnts[1];	
	
	ImageWarpLinearMesh deformQuad(img, img.width(), img.height(), numPnts, inv, true, 1, 0, ImagePenalization(), &hierarchyDescr, FACE_QUAD);
	deformQuad.Deform(deformQuad.GetParameterVector());
	deformQuad.m_bShowWireframe = wireframe;
	deformQuad.GetDeformedImage().save("ImgWarpUnmodified.png");	
	std::cout<<"Hierarchies="<<deformQuad.m_iMaxHierarchyLevels<<std::endl;
	//deform deform image low-level hierarchy
	deformQuad.SwitchToHierarchy(0);
	SymVectorReal & paramsLow = deformQuad.GetParameterVector();	
	std::cout<<"Low-level: "<<paramsLow.m<<std::endl;
	int lowParamID = ((lowResDim-1)*lowResDim + lowResDim-1)*2;
	paramsLow(lowParamID) += rndLevelLow;		paramsLow(lowParamID+1) += rndLevelLow;
	((SymMatrix2DReal &) deformQuad.DeformAtCurrentHierarchyLevel(paramsLow)).ToImage().save("ImageWarp1LowF.png");
	deformQuad.GetDeformedImage().save("ImgWarp1Low.png");
	paramsLow(lowParamID) -= rndLevelLow;		paramsLow(lowParamID+1) -= rndLevelLow;	
	
	//deform deform image mid-level hierarchy
	deformQuad.SwitchToHierarchy(1);	
	SymVectorReal & paramsMed = deformQuad.GetParameterVector();		
	std::cout<<"Mid-level: "<<paramsMed.m<<std::endl;	
	int midParamID = ((midResDim-1)*midResDim + midResDim-1)*2;	
	paramsMed(midParamID) += rndLevelMed;		paramsMed(midParamID+1) += rndLevelMed;	
	((SymMatrix2DReal &) deformQuad.DeformAtCurrentHierarchyLevel(paramsMed)).ToImage().save("ImageWarp2MedF.png");
	deformQuad.GetDeformedImage().save("ImgWarp2Med.png");
	paramsMed(midParamID) -= rndLevelMed;		paramsMed(midParamID+1) -= rndLevelMed;		
	
	//deform deform image top-level hierarchy	
	deformQuad.SwitchToHierarchy(2);	
	SymVectorReal & paramsHigh = deformQuad.GetParameterVector();	
	std::cout<<"High-level: "<<paramsHigh.m<<std::endl;		
	int highParamID = ((highResDim-1)*highResDim + highResDim-1)*2;	
	paramsHigh(highParamID) += rndLevelHigh;		paramsHigh(highParamID+1) += rndLevelHigh;	
	((SymMatrix2DReal &) deformQuad.DeformAtCurrentHierarchyLevel(paramsHigh)).ToImage().save("ImageWarp3HighF.png");
	deformQuad.GetDeformedImage().save("ImageWarp3High.png");
	paramsHigh(highParamID) -= rndLevelHigh;		paramsHigh(highParamID+1) -= rndLevelHigh;	
	
//	//deform combined
//	params(targetHighLevel) += rndLevelHigh;	params(targetHighLevel+1) += rndLevelHigh;	
//	params(targetMidLevel) += rndLevelMed;		params(targetMidLevel+1) += rndLevelMed;	
//	params(targetLowLevel) += rndLevelLow;		params(targetLowLevel+1) += rndLevelLow;	
//	((SymMatrix2DReal &) deformQuad.DeformAtCurrentHierarchyLevel(params)).ToImage().save("ImageWarp4CombF.png");
//	deformQuad.GetDeformedImage().save("ImgWarp4Comb.png");	
}

void ImageWarpLinearMesh::TestImageWarpFoldover()
{
	//generate image without foldover (see energy)
	//prepare image
	QImage img = TestCreateCheckerboard(8, 8, 50, 50, 0);
	
	int numPnts[2];
	numPnts[0] = 1;
	numPnts[1] = 1;	
	bool inv = true;
	bool fixBnd = true;
	double disp = 120;
	if (inv)
		disp = disp/qMax(img.width(), img.height());
	bool wireframe = true;
	double penalty = 0;
	ImagePenalization penalization;
	penalization.m_iFields = ImageWarp::ImagePenalization::FOLDOVER_BINARY;
	
	//generate image with foldover (see energy)
	//tri1
	ImageWarpLinearMesh deformTri1(img, img.width(), img.height(), numPnts, inv, fixBnd, 1, 0, penalization, NULL, FACE_TRIANGLE1);	
	deformTri1.m_bShowWireframe = wireframe;
	SymVectorReal & params = deformTri1.GetParameterVector();	
	
	params(0) -= disp/2;		params(1) -= disp/2;			
	((SymMatrix2DReal&)deformTri1.DeformAtCurrentHierarchyLevel(params, &penalty));
	deformTri1.GetDeformedImage().save("PenaltyTestT1_1.png");
	std::cout<<"PenaltyTestT1_1.png: penalty="<<penalty<<std::endl;
		
	params(0) -= disp/2;		params(1) -= disp/2;	
	((SymMatrix2DReal&)deformTri1.DeformAtCurrentHierarchyLevel(params, &penalty));
	deformTri1.GetDeformedImage().save("PenaltyTestT1_2.png");
	std::cout<<"PenaltyTestT1_2.png: penalty="<<penalty<<std::endl;	
	
	//tri2	
	ImageWarpLinearMesh deformTri2(img, img.width(), img.height(), numPnts, inv, fixBnd, 1, 0, penalization, NULL, FACE_TRIANGLE2);	
	deformTri2.m_bShowWireframe = wireframe;
	params = deformTri2.GetParameterVector();	
	
	params(0) -= disp/2;		params(1) += disp/2;			
	((SymMatrix2DReal&)deformTri2.DeformAtCurrentHierarchyLevel(params, &penalty));
	deformTri2.GetDeformedImage().save("PenaltyTestT2_1.png");
	std::cout<<"PenaltyTestT2_1.png: penalty="<<penalty<<std::endl;
	
	params(0) -= disp/2;		params(1) += disp/2;	
	((SymMatrix2DReal&)deformTri2.DeformAtCurrentHierarchyLevel(params, &penalty));
	deformTri2.GetDeformedImage().save("PenaltyTestT2_2.png");
	std::cout<<"PenaltyTestT2_2.png: penalty="<<penalty<<std::endl;	
	
	//tri3	
	ImageWarpLinearMesh deformTri3(img, img.width(), img.height(), numPnts, inv, fixBnd, 1, 0, penalization, NULL, FACE_TRIANGLE3);	
	deformTri3.m_bShowWireframe = wireframe;
	params = deformTri3.GetParameterVector();	
	
	params(0) -= disp/2;		params(1) += disp/2;			
	((SymMatrix2DReal&)deformTri3.DeformAtCurrentHierarchyLevel(params, &penalty));
	deformTri3.GetDeformedImage().save("PenaltyTestT3_1.png");
	std::cout<<"PenaltyTestT3_1.png: penalty="<<penalty<<std::endl;
	
	params(0) -= disp/2;		params(1) += disp/2;	
	((SymMatrix2DReal&)deformTri3.DeformAtCurrentHierarchyLevel(params, &penalty));
	deformTri3.GetDeformedImage().save("PenaltyTestT3_2.png");
	std::cout<<"PenaltyTestT3_2.png: penalty="<<penalty<<std::endl;
	
	params(0) += disp;		params(1) -= disp;	
	params(0) -= disp/2;	params(1) -= disp/2;	
	((SymMatrix2DReal&)deformTri3.DeformAtCurrentHierarchyLevel(params, &penalty));
	deformTri3.GetDeformedImage().save("PenaltyTestT3_3.png");
	std::cout<<"PenaltyTestT3_3.png: penalty="<<penalty<<std::endl;

	params(0) -= disp/2;		params(1) -= disp/2;	
	((SymMatrix2DReal&)deformTri3.DeformAtCurrentHierarchyLevel(params, &penalty));
	deformTri3.GetDeformedImage().save("PenaltyTestT3_4.png");
	std::cout<<"PenaltyTestT3_4.png: penalty="<<penalty<<std::endl;

	
	//quad	
	ImageWarpLinearMesh deformQuad(img, img.width(), img.height(), numPnts, inv, fixBnd, 1, 0, penalization, NULL, FACE_QUAD);	
	deformQuad.m_bShowWireframe = wireframe;
	params = deformQuad.GetParameterVector();	
	
	params(0) -= disp/2;		params(1) += disp/2;			
	((SymMatrix2DReal&)deformQuad.DeformAtCurrentHierarchyLevel(params, &penalty));
	deformQuad.GetDeformedImage().save("PenaltyTestQ_1.png");
	std::cout<<"PenaltyTestQ_1.png: penalty="<<penalty<<std::endl;
	
	params(0) -= disp/2;		params(1) += disp/2;	
	((SymMatrix2DReal&)deformQuad.DeformAtCurrentHierarchyLevel(params, &penalty));
	deformQuad.GetDeformedImage().save("PenaltyTestQ_2.png");
	std::cout<<"PenaltyTestQ_2.png: penalty="<<penalty<<std::endl;	
	
	//smooth energy (i.e. approach foldover, print energy over 10 iterations)
}

void ImageWarpLinearMesh::TestAsRigidAsPossibleWarp()
{
	QImage img = TestCreateCheckerboard(8, 8, 50, 50, 0);
	int * numPnts = new int[2];
	numPnts[0] = 20;
	numPnts[1] = 20;	
	ImageHierarchyDescriptor * hierarchyDescr = new ImageHierarchyDescriptor();
	hierarchyDescr->m_HierarchyType = AS_RIGID_AS_POSSIBLE;
	hierarchyDescr->m_iMinPnts = 2;
	hierarchyDescr->m_iMaxPnts = 4;
	ImageWarpLinearMesh asRigidMesh(img, 250, 250, numPnts, false, false, 10, 0.1, ImagePenalization(), hierarchyDescr);
	
		double cx, cy;
		asRigidMesh.GetCenterOfMass(&cx, &cy);
	std::cout<<"INITIAL PARAMS= "<<std::endl; asRigidMesh.GetParameterVector().PrintVector(); std::cout<<std::endl;	
	std::cout<<"cx="<<cx<<" cy="<<cy<<std::endl;
		asRigidMesh.SetToPolarCoords(cx, cy);
	
	std::cout<<"FINAL PARAMS= "<<std::endl; asRigidMesh.GetParameterVector().PrintVector(); std::cout<<std::endl;		
	
	asRigidMesh.m_bShowWireframe = true;
	asRigidMesh.m_bPreserveScale = true;
	std::cout<<"Rigid Hierarchies = "<<asRigidMesh.m_iMaxHierarchyLevels<<std::endl;
	asRigidMesh.SwitchToHierarchy(0);
	std::cout<<"Switch Successful"<<std::endl;	
	SymVectorReal & params = asRigidMesh.GetParameterVector();
	double moveByFrac = 0.4;
	// Cartesian
//	params(0) += (params(2) - params(0)) * moveByFrac;
//	params(1) += (params(3) - params(1)) * moveByFrac;
//	params(2) -= (params(2) - params(0)) * moveByFrac;
//	params(3) -= (params(3) - params(1)) * moveByFrac;
//
//	params(4) += (params(2) - params(4)) * moveByFrac;
//	params(5) += (params(3) - params(5)) * moveByFrac;
//	//	params(6) += (params(2) - params(6)) * moveByFrac;
//	params(7) += (params(3) - params(7)) * moveByFrac*2;

	params(0) -= 125;
	params(1) += 3.14/2;
	params(2) = 0;
	//	params(3) += 3.14 / 20;
	
	
	((SymMatrix2DReal&)asRigidMesh.DeformAtCurrentHierarchyLevel(params)).ToImage().save("NotFixedBoundary.png");
	std::cout<<"Deformation successful"<<std::endl;		
	asRigidMesh.GetDeformedImage().save("AsRigidTest_Step12.png");
	asRigidMesh.GetDeformedImage(ImageWarpLinearMesh::IMAGE_TYPE_TEXTURE_SPACE).save("AsRigidTest_Step12_TextureCoord.png");
	asRigidMesh.DrawGeodesicSorting().save("GeodesicSorting_Step12.png");
	asRigidMesh.PrintParams(true);

	asRigidMesh.m_bPreserveScale = false;
	asRigidMesh.DeformAtCurrentHierarchyLevel(params);
	asRigidMesh.GetDeformedImage().save("AsRigidTest_Step1Only.png");
	asRigidMesh.DrawGeodesicSorting().save("GeodesicSorting_Step1Only.png");
	
	asRigidMesh.GetDeformedImage(-1).save("AsRigidTest_Final.png");	

	asRigidMesh.SwitchToHierarchy(0);	
	std::cout<<"Hierarchy 0: Suggested = ";		asRigidMesh.GetSuggestedParamStep().PrintVector(0.001);		std::cout<<std::endl;
	std::cout<<"             Minimal   = ";		asRigidMesh.GetMinimalParamStep().PrintVector(0.001);		std::cout<<std::endl;	

	asRigidMesh.SwitchToHierarchy(1);	
	std::cout<<"Hierarchy 1: Suggested = ";		asRigidMesh.GetSuggestedParamStep().PrintVector(0.001);		std::cout<<std::endl;
	std::cout<<"             Minimal   = ";		asRigidMesh.GetMinimalParamStep().PrintVector(0.001);		std::cout<<std::endl;	
}

void ImageWarpLinearMesh::TestAsRigidAsPossibleWarpHierarchySwitchConsistency()
{
	QImage img = TestCreateCheckerboard(8, 8, 50, 50, 0);
	int * numPnts = new int[2];
	numPnts[0] = 20;
	numPnts[1] = 20;	
	ImageHierarchyDescriptor * hierarchyDescr = new ImageHierarchyDescriptor();
	hierarchyDescr->m_HierarchyType = AS_RIGID_AS_POSSIBLE;
	hierarchyDescr->m_iMinPnts = 2;
	hierarchyDescr->m_iMaxPnts = 4;
	ImageWarpLinearMesh asRigidMesh(img, img.width(), img.height(), numPnts, false, false, 10, 0.1, ImagePenalization(), hierarchyDescr);
	asRigidMesh.m_bShowWireframe = true;
	std::cout<<"Rigid Hierarchies = "<<asRigidMesh.m_iMaxHierarchyLevels<<std::endl;
	
	asRigidMesh.SwitchToHierarchy(0);
	SymVectorReal & params = asRigidMesh.GetParameterVector();
	double moveByFrac = 0.4;
	params(0) += (params(2)-params(0)) * moveByFrac;
	params(1) += (params(3)-params(1)) * moveByFrac;
	params(2) += (params(0)-params(2)) * moveByFrac;
	params(3) += (params(1)-params(3)) * moveByFrac;	
	asRigidMesh.DeformAtCurrentHierarchyLevel(params);
	asRigidMesh.GetDeformedImage().save("AsRigidTest_2pnts.png");
	asRigidMesh.SwitchToHierarchy(1);
	asRigidMesh.DeformAtCurrentHierarchyLevel(asRigidMesh.GetParameterVector());
	asRigidMesh.GetDeformedImage().save("AsRigidTest_4pnts.png");	
}

void ImageWarpLinearMesh::TestFeatureBasedMeshing()
{
	std::cout<<"Starting test"<<std::endl;
//	QImage img = TestCreateCheckerboard(8, 8, 50, 50, 0);
//	QImage img("../../../data/imageSymmetrization/1.161_w17x17.jpg");
	QImage img("../../../data/imageSymmetrization/3.bricks_sm7x11.jpg");	
	int numPnts = 400;
	ImageHierarchyDescriptor * hierarchyDescr = new ImageHierarchyDescriptor();	
	hierarchyDescr->m_HierarchyType = AS_RIGID_AS_POSSIBLE;
	hierarchyDescr->m_iMinPnts = 2;
	hierarchyDescr->m_iMaxPnts = 4;
	
	ImageWarpLinearMesh arbitraryMesh(img, img.width(), img.height(), numPnts, false, false, 10, 0, ImagePenalization(), hierarchyDescr);
	arbitraryMesh.SwitchToHierarchy(1);
	arbitraryMesh.m_bShowWireframe = true;
	arbitraryMesh.DeformAtCurrentHierarchyLevel(arbitraryMesh.GetParameterVector());
	arbitraryMesh.DrawGeodesicSorting().save("FeatureBasedMeshSorting.png");
	arbitraryMesh.GetDeformedImage().save("FeatureBasedMesh0.png");
	std::cout<<"Foldover penalty = "<<arbitraryMesh.CalculateDeformationPenalty()<<std::endl;
	SymVectorReal & params = arbitraryMesh.GetParameterVector();
	params(0) += 30;
	params(1) += 30;	
	params(2) += 30;
	params(3) -= 30;	
	params(4) -= 25;
	params(5) -= 50;	
	
	arbitraryMesh.DeformAtCurrentHierarchyLevel(params);
	arbitraryMesh.GetDeformedImage().save("FeatureBasedMesh1.png");	
}





