#include "Numerical.h"
#include "SymFilters.h"

#ifndef __DEFORMATION_H
#define __DEFORMATION_H

class DeformationModel;

class ExampleWithAnswer
{
	public:
		ExampleWithAnswer(QString name);
		double GetParameterSpaceError(SymVectorReal * optimizerOutput);
	
		DeformationModel * m_pValidSolutionSpace;
		SymVectorReal * m_pPerfectSolution;

		QString m_sName;
	
		// if image is among outputs
		double GetImageSpaceError(QImage & optimizerSolution);
		double GetSymSpaceError(SymMatrix & optimizerSolution);
		
		
		QImage m_ImageInputExample;
		QImage m_ImagePerfectOutput;
		QImage m_ImagePerfectDeformed;
		SymMatrix * m_pPerfectSymmetrySpace;
	
		double m_fEnergyPerfect;
		double m_fEnergyDeform;
	
};

class DeformationHierarchyNode
{
	public:
		DeformationHierarchyNode();
		enum HierarchyRelationship
		{
			HIERARCH_REL_ADD,
			HIERARCH_REL_OFFSET_FROM_AVERAGE
		};
	
		std::vector<int> m_ParamIDs;
		std::vector<std::vector<int> > m_LevelParamStepIDs;
		std::vector<DeformationHierarchyNode *> m_apChildren;
		std::vector<double> m_afChildrenWeights;
		std::vector<DeformationHierarchyNode *> m_apParents;	
		std::vector<int> m_iIDinParents;
		HierarchyRelationship m_RelationshipToParent;
		double m_fTempWeightValue;
		bool m_bAdded;
		int m_iDepth;
};

//If you plan to extend hierarchy level to include multiple levels
//derive subclass for many hierarchies, with similar functions
class HierarchyLevelDescriptor
{
	public:	
		HierarchyLevelDescriptor(int depth, int firstElement, int maxElements, 
								 std::vector<DeformationHierarchyNode*> & breadthFirstSorted);
		virtual ~HierarchyLevelDescriptor(){}
	
		virtual void CopyFromConcatParams(SymVectorReal * trgParams);
		virtual void CopyFromConcatSteps(SymVectorReal * trgParams);
		virtual void CopyFromConcatMinSteps(SymVectorReal * trgParams);
	
		virtual void CopyToConcatParams(SymVectorReal * srcParams);
		virtual void CopyToConcatSteps(SymVectorReal * srcParams);
		virtual void CopyToConcatMinSteps(SymVectorReal * srcParams);

		SymVectorReal * m_pConcatSteps;
		SymVectorReal * m_pConcatParams;
		SymVectorReal * m_pConcatMinSteps;	
		int * m_aiMapToConcatParams;
		int * m_aiMapToConcatParamsSteps;
		
		int m_iDepth;
		
		bool m_bSeparateStepMapping;
};

class DeformationModel
{
	public:	
		enum OutputType
		{
			DEFORM_OUT_IMG2D,
			DEFORM_OUT_SYMREAL2D
		};
	
		enum HierarchyTreeTraversal
		{
			HIERARCHY_TRAVERSE_DEPTH,
			HIERARCHY_TRAVERSE_BREADTH,
			HIERARCHY_TRAVERSE_NONE
		};
	
		enum DeduceWeightsMethod
		{
			WEIGHTS_KEEP_AS_IT_IS,		/** Do not change weights */
			WEIGHTS_CONST,				/** Have all children weight 1 */
			WEIGHTS_DISTANCE_LINEAR,	/** Sets weights proportional to 2*MAX-Dist */
			WEIGHTS_BILINEAR,			/** bilinear weights, dimensionality decided in child */
			WEIGHTS_BSPLINE				/** b-spline weights, dimensionality decided in child */
		};
	
		enum DeduceParentsLocations
		{
			PARENTSLOC_KEEP_AS_IT_IS,		/** Do not change parent's location */
			PARENTSLOC_AVERAGE_CHILDREN		/** Set parent's location to average of its children */
		};
		enum DeformationResults
		{
			DEFORMATION_FINAL_RESULT = -1,
		};
	
		DeformationModel();
		virtual ~DeformationModel();
	
		virtual bool DistortsInput();
		virtual void InvertColors();
	
		virtual void GetOutputDimensions(int * inWidth, int * inHeight);
		virtual void GetCenterOfMass(double * cx, double * cy);
		virtual DeformationModel * CloneSelfUsingBestParams(SymVectorReal & bestParams);
		
		virtual void ChildUpdateAtHierarchySwitch(int currLevel, int newLevel);
		virtual SymMatrix & DeformAtCurrentHierarchyLevel(SymVectorReal & params, double * outPenalty = NULL);	
		virtual SymMatrix & Deform(SymVectorReal & params, double * outPenalty = NULL) = 0;
		virtual void SetPostFilter(SymFilter * filter);
		
		virtual bool HasOutput(OutputType type);
		virtual QImage GetDeformedImage(int imageType = 0);
		virtual int GetNumOutputImageTypes();
	
		virtual int GetNumHierarchyLevels();
		virtual bool SwitchToHierarchy(int hierarchy);
		virtual int GetCurrentHierarchy();
		virtual bool GoUpHieararchy();
		virtual bool GoDownHierarchy();
	
		virtual void PrintParams(bool printMin);
		virtual SymVectorReal & GetParameterVector();
		virtual SymVectorReal & GetSuggestedParamStep();
		virtual SymVectorReal & GetMinimalParamStep();	
		virtual void ReduceSuggestedStepByFactor(double factor);
		virtual void TestRandomizeParams(double minRand, double maxRand);
	
		virtual SymMatrix & TakeStepInDimension(int dimension, double dx, double & penalty);
		virtual void RevertLastStepInDimension();
	
		virtual SymMatrix & TakeRandomStep(bool & enforceStep);
		virtual bool TerminalState();
		virtual void RevertLastRandomStep();
		int m_iLastChangedParam;
		double m_fLastParamChangedValue;
	
		HierarchyTreeTraversal m_TraversalType;
	
		void SetToPolarCoords(double centerX, double centerY);
	
	
	/**
	 * Initializes hierarchy, optinal: fills initial parameters (i.e. sets all offsets for higher-level params)
	 * Note this is support function, intended to be used by subclass.
	 * @param hierarchyTreeRoots	hierarchy trees must be generated in advance (e.g. multi-res grid for images)
	 * @param params				parameters for each point in hierarchy (at least at the finest resolution). 
	 * @param traversalType			keep it BREADTH_FIRST for now, in general there could be different ways to traverse the tree
	 * @param weightingMethod		how children weights are estimated (unless provided via m_afChildrenWeights)
	 * @param pLocMethod			how parent's 'initial parameter' is estimated (unless provided via params)
	 */
		void InitializeParameterHierearchy(std::vector<DeformationHierarchyNode *> &hierarchyTreeRoots, 
										   SymVectorReal & params, HierarchyTreeTraversal traversalType, 
										   DeduceWeightsMethod weightingMethod, DeduceParentsLocations pLocMethod);
	/**
	 * DON'T use me! use breadth-first option instead.
	 */
	//	double RecursivelyFillValue(DeformationHierarchyNode * root, SymVectorReal & params);
	
	/**
	 * Fills initial paramaters, weights, etc...  in a breadth-first manner. See InitializeParameterHierearchy
	 */
		void BreadthFirstFillValues(SymVectorReal & params, DeduceWeightsMethod weightingMethod, 
									DeduceParentsLocations pLocMethod);
	
	/**
	 * fills paramter values according to hierarchy
	 */
		void GenerateHierarchicalParameterVector(SymVectorReal & params);
		virtual void GenerateHierarchicalParameterVectorChildImplement(SymVectorReal & params, SymVectorReal & trg);
	
	/**
	 * DON'T use me! use breadth-first option instead.
	 */	
	//	void DepthFirstHierarchicalParameterVector(DeformationHierarchyNode * root);
	
	/**
	 * Fills parameters according to hierarchy.
	 */
		void BreadthFirstHierarchicalParameterVector();
		
	protected:
		bool m_bInvertColors;
		bool m_bPolarCoords;
		double m_fPolarCoordsCX;
		double m_fPolarCoordsCY;
		void ConvertCart2Polar(SymVectorReal & vecFrom, SymVectorReal & vecTo);
		void ConvertPolar2Cart(SymVectorReal & vecFrom, SymVectorReal & vecTo);	
	
	
		struct HierarchyPrintNode
		{
			HierarchyPrintNode(QString name, std::vector<int> &paramIDs, int depth);
			int m_iDepth;
			QString m_sName;
			std::vector<int> m_iParamIDs;
		};
		std::vector<HierarchyPrintNode> m_ParamPrintNodes;	
	
		SymVectorReal * m_pInitParams;
		SymVectorReal * m_pParameterSteps;
		SymVectorReal * m_pMinimalParamSteps;

		SymVectorReal * m_pAllLastParams;	//in child dimensions
		SymVectorReal * m_pAllParamStep;
		SymVectorReal * m_pAllMinParamStep;
	
		SymVectorReal * m_pFinalParams;
		
		SymMatrix * m_pDeformedMatrix;
	
		std::vector<DeformationHierarchyNode *> m_pHierarchyTreeRoots;
		std::vector<DeformationHierarchyNode *> m_pHierarchyBreadthFirstList;
	
		std::vector<HierarchyLevelDescriptor * > m_PrecomputedSingleLayers;
		std::vector<int> m_iLayerToFirstElement;
		std::vector<int> m_iLayerToMaxElements;
	
		int m_iHierarchyLevel;
		int m_iMaxHierarchyLevels;
	
		SymFilter * m_pPostFilter;
		
};

#endif




