#include "Deformations.h"
#include "SymEnergy.h"
#include "Numerical.h"
#include <gsl/gsl_multimin.h>
#include <Qt/qmap.h>
#include "TimeProfiler.h"

#ifndef __SYM_OPTIMIZER_H
#define __SYM_OPTIMIZER_H


//this function should return the result f(x,params) for argument x and parameters params. 
//If the function cannot be computed, an error value of GSL_NAN should be returned.
double optimF(const gsl_vector * x, void * params);
//this function should store the n-dimensional gradient g_i = d f(x,params) / d x_i in the vector 
//g for argument x and parameters params, returning an appropriate error code if the function cannot be computed.
void optimDF(const gsl_vector * x, void * params, gsl_vector * g);
//This function should set the values of the f and g as above, for arguments x and parameters params. 
//This function provides an optimization of the separate functions for f(x) and g(x)—it is always faster 
//to compute the function and its derivative at the same time. 
void optimFDF(const gsl_vector * x, void * params, double * f, gsl_vector * g);

class SymOptimizer
{
	public:
		enum OptimizationAlgirithm
		{
			FDF_CONJ_FLETCHER_REEVES,
			FDF_CONJ_POLAK_RIBIERE,
			FDF_BFGS,				// choose me if gradients are smooth.
			FDF_GRAD_DESC,
			F_SIMPLEX_NELDER_MEAD,	//choose me if no gradients are availiable
			GREEDY_WALK,			//for texture synth
			MY_GRADIENT_DESCENT		// gradient descent one variable at a time
		};
	
		enum IterationType
		{
			ITERATE_BY_FN_EVALS,
			ITERATE_BY_ALGORITHM_ITERS,
			ITERATE_BY_TIME
		};
	
		SymOptimizer(SymEnergy * energyFn, DeformationModel * deformationModel, int maxIterations, 
					 int maxFunctionEvals, int updateAfterIterations, double rescaleAtUpdate,
					 ExampleWithAnswer * perfectSolution = NULL);
		virtual ~SymOptimizer(){}
		virtual void Iterate() = 0;
		virtual SymVectorReal * GetBestSolution();
		
		virtual void StartIterating();
		virtual bool KeepIterating(int maxIters);
		virtual bool RequireUpdate(double bestEnergy, double currentStepChange, bool saveHistoryLandmarkAndRescale);
		virtual bool CheckSuccess();
		virtual void PrintUpdate();
		
		virtual double ForwardDifference(gsl_vector * tmpVec, int dim, double dx, double currVal);
		virtual double CentralDifference(gsl_vector * tmpVec, int dim, double dx);
		gsl_vector * m_pTempVec;	
		double m_fStep;
		int m_iDerivativeEvals;		

		SymMatrix * m_OptSolution;
		
		std::vector<double> m_EnergyHistory;
		std::vector<double> m_DXStepHistory;
		std::vector<int> m_iFnEvalsHistory;
		std::vector<int> m_HistoryLandmark;
	
		SymEnergy * m_pEnergyFn;
		DeformationModel * m_pDeformationModel;
	
		IterationType m_IterationType;
		int m_iFunctionEvals;
		int m_iNumIterations;
		int m_iTimeStampFnEvals;
		int m_iTimeStampIterations;
		double m_fTimeStampBestValue;
		double m_fParamChangeLastIter;
	
		int m_iMaxIterations;
		int m_iMaxFunctionEvals;
		bool m_bVerbous;
		int m_iUpdateEvery;
		double m_fRescaleAtUpdate;
		TimeProfiler m_Profiler;
		
		SymVectorReal * m_pBestSolution;
		double m_fBestValue;
		int m_iOutputBias;
		bool m_bStepIsTooSmall;

		ExampleWithAnswer * m_pAnswer;
		std::vector<double> m_ParamErrors;
		std::vector<double> m_ImageErrors;
};

class SymOptimizerGSLSmooth : public SymOptimizer
{
	public:
		SymOptimizerGSLSmooth(int numVariables, double step, double minLineTol, double minStepNorm, 
							  SymEnergy * energyFn, DeformationModel * deformationModel, 
							  OptimizationAlgirithm algorithm, int maxIterations, int maxFunctionEvals,
							  int updateAfterIterations, double rescaleAtUpdate, 
							  ExampleWithAnswer * perfectSolution = NULL);
		virtual void Iterate();
	
		gsl_multimin_fdfminimizer * m_pFDFMinimizer;
		OptimizationAlgirithm m_Algorithm;
		int m_iNumVars;	
		double m_fMinStep;
		
		double m_fMinLineTol;
};

class SymOptimizerGSLDiscrete : public SymOptimizer
{
	public:
		SymOptimizerGSLDiscrete(int numVariables, SymEnergy * energyFn, DeformationModel * deformationModel, 
								int maxIterations, int maxFunctionEvals, int updateAfterIterations, 
								double rescaleAtUpdate, ExampleWithAnswer * perfectSolution = NULL);
	
		virtual void Iterate();
	
		double m_fMinStep;	
		gsl_multimin_fminimizer * m_pFMinimizer;
		int m_iNumVars;	
};

class SymOptimizerGreedyWalk : public SymOptimizer
{
	public:
		SymOptimizerGreedyWalk(SymEnergy * energyFn, DeformationModel * deformationModel, 
						int maxIterations, int maxFunctionEvals, int updateAfterIterations, 
						double rescaleAtUpdate, ExampleWithAnswer * perfectSolution = NULL);
		virtual void Iterate();
		bool m_bNoOpimization;
};

class SymOptimizerNDGradientDescent : public SymOptimizer
{
	public:
		SymOptimizerNDGradientDescent(int numDims, SymEnergy * energyFn, DeformationModel * deformationModel, 
									int maxIterations, int maxFunctionEvals, int updateAfterIterations, 
									double rescaleAtUpdate, ExampleWithAnswer * perfectSolution = NULL);
		virtual void Iterate();
	
		static int GetParameterID(int numVars, int grouped, int parameterID, int dim);
	
		static void FindGradientDirection(int numDims, int grouped, int parameterID, SymEnergy * energyFn, 
										  DeformationModel * deformationModel, SymVectorReal & gradDirection,
										  double * optionalScale, double currVal);
		SymVectorReal * m_pGradDirection;
		int m_iNumDimensions;
		int m_iGroupedDims;
};


class MultiLevelOptimizer : public SymOptimizer
{
	public:
		MultiLevelOptimizer(SymEnergy * energyFn, DeformationModel * deformationModel, OptimizationAlgirithm algorithm, 
							int maxIterations, int maxFunctionEvals, int updateAfterIterations, 
							double rescaleAtUpdate, bool refeedImages, ExampleWithAnswer * perfectSolution = NULL,
							int numDims=2);
		virtual void Iterate();
	
		bool m_bRefeedImageAtEachIteration;
	
		std::vector<SymOptimizer*> m_OptimizersPerLevel;
		OptimizationAlgirithm m_Algorithm;
		int m_iNumLevels;
};


#endif

